@@ -8,6 +8,7 @@
 #ifndef STEVE_H
 #define STEVE_H 1
 
+#include <stdint.h>
 #include <sys/ioctl.h>
 
 /* get number of currently available tokens */
@@ -27,6 +28,9 @@
 /* get/set per-process limit */
 #define STEVE_IOC_GET_PER_PROCESS_LIMIT _IOR(0x53, 0x05, int64_t)
 #define STEVE_IOC_SET_PER_PROCESS_LIMIT _IOW(0x53, 0x05, int64_t)
+/* get/set min. memory available (in MiB) */
+#define STEVE_IOC_GET_MIN_MEMORY_AVAIL _IOR(0x53, 0x06, int64_t)
+#define STEVE_IOC_SET_MIN_MEMORY_AVAIL _IOW(0x53, 0x06, int64_t)
 
 #define STEVE_IOC_MASK_GET _IOR(0x53, 0, int64_t)
 #define STEVE_IOC_MASK_SET _IOW(0x53, 0, int64_t)
